import 'dart:async';
import 'dart:convert';

import 'package:bea_flutter_package/bea_services/dtos/bea_models/base_models/error/bea_error.dart';
import 'package:bea_flutter_package/bea_services/dtos/bea_models/base_models/login/bea_impersonate_input.dart';
import 'package:bea_flutter_package/bea_services/dtos/flutter_models/input/login_input.dart';
import 'package:bea_flutter_package/bea_services/dtos/flutter_models/output/bea_session.dart';
import 'package:bea_flutter_package/bea_services/dtos/flutter_models/output/impersonate_output.dart';
import 'package:bea_flutter_package/bea_services/dtos/flutter_models/output/login_output.dart';
import 'package:bea_flutter_package/bea_services/utils/key_utils.dart';
import 'package:bea_flutter_package/bea_services/utils/url_utils.dart';
import 'package:dio/dio.dart';
import 'package:f_logs/model/flog/flog.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart';

import '../dtos/flutter_models/output/generic_server_response.dart';
import 'bea_service_manager.dart';

class BeaAccountManager{
  final BeaServiceManager bsm;

  String _baseUrl = "";
  FlutterSecureStorage storage;

  BeaAccountManager(this.bsm){
    try{
      if(bsm == null){
        throw new Exception("${this.runtimeType.toString()} not initialized");
      }

      _baseUrl = bsm.getBaseUrl();
      storage = bsm.getBeaStorage();
    }on Exception catch (e){
      print('Initialization ${this.runtimeType.toString()} error');
      FLog.error(
          className: this.runtimeType.toString(),
          methodName: "Costructor",
          text: 'Initialization ${this.runtimeType.toString()} error - $e');
      throw e;
    }
  }

  Future<GenericServerResponse<LoginOutput>> login(LoginInput input) async {

    GenericServerResponse<LoginOutput> genericServerResponse = new GenericServerResponse();
    genericServerResponse.data = new LoginOutput();

    try{

      String url = UrlUtils.getUrl(_baseUrl, UrlUtils.URL_AUTH);

      String jsonInput = jsonEncode(input.toJson());

      Response response = await bsm.genericBeaHttpPost(url, jsonInput);
      int statusCode = response.statusCode;

      if(statusCode == 200){
        String token = response.data['result'];

        if(token == null || token.isEmpty){
          throw new Exception('Impossibile reperire il token');
        }

        await storage.write(key: KeyUtils.KEY_RICORDAMI, value: input.ricordami == true ? "true" : "false");
        await storage.write(key: KeyUtils.KEY_TOKEN_SESSION, value: token);
        await storage.write(key: KeyUtils.KEY_USER, value: input.usernameOrEmailAddress);

        if(input.ricordami){
          await storage.write(key: KeyUtils.KEY_PASSWORD, value: input.password);
        }

        genericServerResponse.data.user = input.usernameOrEmailAddress;
        genericServerResponse.data.token = token;
        genericServerResponse.data.isLogged = true;

      }else if(statusCode == 500){
        BeaError error = BeaError.fromJson(response.data['error']);
        if (error != null) {
          genericServerResponse.data.user = input.usernameOrEmailAddress;
          genericServerResponse.data.token = "";
          genericServerResponse.data.isLogged = false;
          genericServerResponse.beaError = error;
        } else {
          genericServerResponse.beaError = new BeaError("Unknown Error", 0, "");
        }
      }else{
        genericServerResponse.beaError = new BeaError('Codice $statusCode ricevuto dal server non gestito!', 0, "");
        throw new Exception('Codice $statusCode ricevuto dal server non gestito!');
      }
    }on Exception catch (e){
      print('BEA Login error');
      FLog.error(
          className: this.runtimeType.toString(),
          methodName: "login",
          text: 'BEA Login error - $e');
      throw e;
    }

    return genericServerResponse;
  }

  Future<GenericServerResponse<LoginOutput>> refreshToken() async {

    GenericServerResponse<LoginOutput> genericServerResponse = new GenericServerResponse<LoginOutput>();

    try{

      String ric = await storage.read(key: KeyUtils.KEY_RICORDAMI);
      bool ricordami = ric.toLowerCase() == "true" ? true : false;
      String user = await storage.read(key: KeyUtils.KEY_USER);
      String password = await storage.read(key: KeyUtils.KEY_PASSWORD);

      if(ricordami != null && ricordami){
        LoginInput loginInput = new LoginInput();
        loginInput.ricordami = ricordami;
        loginInput.tenancyName = "";
        loginInput.usernameOrEmailAddress = user;
        loginInput.password = password;

        genericServerResponse = await login(loginInput);
      }else{
        genericServerResponse.beaError = new BeaError("Session Expired", 401, "Session Expired");
      }
    }on Exception catch (e){
      print('BEA Refresh Token error');
      FLog.error(
          className: this.runtimeType.toString(),
          methodName: "refreshToken",
          text: 'BEA Refresh Token error - $e');
      throw e;
    }

    return genericServerResponse;
  }

  Future logout() async {
    try{
      await storage.deleteAll();
    }on Exception catch (e){
      print('BEA Logout error');
      FLog.error(
          className: this.runtimeType.toString(),
          methodName: "logout",
          text: 'BEA Logout error - $e');
      throw e;
    }
  }

  Future<bool> isLogged() async {
    try{
      String user = await storage.read(key: KeyUtils.KEY_USER);
      if(user != null && user.isNotEmpty){
        return true;
      }else{
        return false;
      }
    }on Exception catch (e){
      print('BEA isLogged error');
      FLog.error(
          className: this.runtimeType.toString(),
          methodName: "isLogged",
          text: 'BEA isLogged error - $e');
      throw e;
    }
  }

  Future<GenericServerResponse<ImpersonateOutput>> impersonate(BeaImpersonateInput input) async {

    GenericServerResponse<ImpersonateOutput> genericServerResponse = new GenericServerResponse();
    genericServerResponse.data = new ImpersonateOutput();

    try{

      String url = UrlUtils.getUrl(_baseUrl, UrlUtils.URL_IMPERSONATE);

      url = "$url?userId=${input.userId}";

      Response response = await bsm.genericBeaHttpPost(url, "");
      int statusCode = response.statusCode;

      if(statusCode == 200){
        String impersonateToken = response.data['result'];

        if(impersonateToken == null || impersonateToken.isEmpty){
          throw new Exception('Impossibile reperire il token di impersonificazione');
        }

        //leggo l'attuale token e lo salvo da parte per poterlo
        //ripristinare dopo
        String currentToken = await storage.read(key: KeyUtils.KEY_TOKEN_SESSION);
        await storage.write(key: KeyUtils.KEY_TOKEN_SESSION_BEFORE_IMPERSONATE, value: currentToken);
        await storage.write(key: KeyUtils.KEY_TOKEN_SESSION, value: impersonateToken);

        genericServerResponse.data.userId = input.userId;
        genericServerResponse.data.token = impersonateToken;
        genericServerResponse.data.isImpersonated = true;

      }else if(statusCode == 500){
        BeaError error = BeaError.fromJson(response.data['error']);
        if (error != null) {
          genericServerResponse.data.userId = input.userId;
          genericServerResponse.data.token = "";
          genericServerResponse.data.isImpersonated = false;
          genericServerResponse.beaError = error;
        } else {
          genericServerResponse.beaError = new BeaError("Unknown Error", 0, "");
        }
      }else{
        genericServerResponse.beaError = new BeaError('Codice $statusCode ricevuto dal server non gestito!', 0, "");
        throw new Exception('Codice $statusCode ricevuto dal server non gestito!');
      }
    }on Exception catch (e){
      print('BEA Impersonate error');
      FLog.error(
          className: this.runtimeType.toString(),
          methodName: "impersonate",
          text: 'BEA Impersonate error - $e');
      throw e;
    }

    return genericServerResponse;
  }

  Future finishImpersonate() async {
    try{
      //ripristino il vecchio token a prima dell'impersonificazione
      String oldToken = await storage.read(key: KeyUtils.KEY_TOKEN_SESSION_BEFORE_IMPERSONATE);
      await storage.write(key: KeyUtils.KEY_TOKEN_SESSION, value: oldToken);
      await storage.delete(key: KeyUtils.KEY_TOKEN_SESSION_BEFORE_IMPERSONATE);
    }on Exception catch (e){
      print('BEA Finish Impersonate error');
      FLog.error(
          className: this.runtimeType.toString(),
          methodName: "finishImpersonate",
          text: 'BEA Finish Impersonate error - $e');
      throw e;
    }
  }

  Future<GenericServerResponse<bool>> isImpersonate() async {

    GenericServerResponse<bool> genericServerResponse = new GenericServerResponse<bool>();

    try{
      String url = UrlUtils.getUrl(_baseUrl, UrlUtils.URL_IS_IMPERSONATOR);

      String jsonInput = '{}';

      Response response = await bsm.genericBeaHttpPost(url, jsonInput);
      int statusCode = response.statusCode;

      switch(statusCode){
        case 200:
          genericServerResponse.data = response.data['result'].toString().toLowerCase() == "true" ? true : false;
          break;
        case 500:
          BeaError error = BeaError.fromJson(response.data['error']);
          if (error != null) {
            genericServerResponse.beaError = error;
          } else {
            genericServerResponse.beaError = new BeaError("Unknown Error", 0, "");
          }
          break;
        case 401:
          genericServerResponse.beaSession = new BeaSession(true);
          break;
      }
    }on Exception catch (e){
      print('BEA is Impersonate error');
      FLog.error(
          className: this.runtimeType.toString(),
          methodName: "isImpersonate",
          text: 'BEA is Impersonate error - $e');
      throw e;
    }

    return genericServerResponse;
  }
}